from pydantic import BaseModel


class ModelUsage(BaseModel):
    input_tokens: int = 0
    output_tokens: int = 0
    total_tokens: int = 0


class InferenceCost(BaseModel):
    input_cost: float = 0.0
    output_cost: float = 0.0

    @property
    def total_cost(self) -> float:
        return self.input_cost + self.output_cost



import json
from pathlib import Path
from typing import Any

from pydantic import BaseModel


def serialize_value(value: Any) -> Any:
    """Convert a Python value to a JSON-serializable format."""
    # Handle None
    if value is None:
        return None

    # Handle primitives
    if isinstance(value, (str, int, float, bool)):
        return value

    # Handle Pydantic models
    if isinstance(value, BaseModel):
        return value.model_dump()

    # Handle lists and tuples
    if isinstance(value, (list, tuple)):
        return [serialize_value(item) for item in value]

    # Handle dicts
    if isinstance(value, dict):
        return {str(k): serialize_value(v) for k, v in value.items()}

    # Fallback: convert to string
    return str(value)


def write_json(data: dict, path: Path) -> None:
    """Write data to a JSON file, creating parent directories if needed."""
    path.parent.mkdir(parents=True, exist_ok=True)
    with open(path, "w") as f:
        json.dump(data, f, indent=2, default=str)



from __future__ import annotations

from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass
from typing import Generic, Iterator, List, TypeVar

T = TypeVar("T")


@dataclass
class ContextStack(Generic[T]):
    """Stack-based context manager for thread-local state"""

    name: str
    _stack: ContextVar[List[T]]

    def __init__(self, name: str, initial: List[T] = []):
        self.name = name
        self._stack = ContextVar(name, default=initial)

    @property
    def current(self) -> T | None:
        """Get current top of stack"""
        stack = self.stack
        return stack[-1] if stack else None

    @property
    def stack(self) -> List[T]:
        return self._stack.get()

    @contextmanager
    def push(self, item: T) -> Iterator[T]:
        token = self._stack.set(self.stack + [item])
        try:
            yield item
        finally:
            self._stack.reset(token)



import asyncio
import logging
import random
from typing import Any, Awaitable, Callable, Type, TypeVar

from pydantic import BaseModel
from tenacity import AsyncRetrying, RetryError, stop_after_attempt, wait_exponential

from agentlens.message import system_message, user_message
from agentlens.provider import Message, Model

logger = logging.getLogger(__name__)

T = TypeVar("T", bound=BaseModel)


async def generate_text(
    model: Model,
    messages: list[Message] | None = None,
    system: str | None = None,
    prompt: str | None = None,
    dedent: bool = True,
    max_retries: int = 3,
    **kwargs,
) -> str:
    return await _generate(
        model.provider.generate_text,
        semaphore=model.provider.get_semaphore(model.name),
        model_name=model.name,
        messages=messages,
        system=system,
        prompt=prompt,
        dedent=dedent,
        max_retries=max_retries,
        **kwargs,
    )


async def generate_object(
    model: Model,
    schema: Type[T],
    messages: list[Message] | None = None,
    system: str | None = None,
    prompt: str | None = None,
    dedent: bool = True,
    max_retries: int = 3,
    **kwargs,
) -> T:
    # inline schemas may have invalid __name__ attributes -- replace w/ a default
    if hasattr(schema, "__name__"):
        schema.__name__ = "Response"
    return await _generate(
        model.provider.generate_object,
        semaphore=model.provider.get_semaphore(model.name),
        model_name=model.name,
        schema=schema,
        messages=messages,
        system=system,
        prompt=prompt,
        dedent=dedent,
        max_retries=max_retries,
        **kwargs,
    )


async def _generate(
    generate: Callable[..., Awaitable[Any]],
    semaphore: asyncio.Semaphore,
    model_name: str,
    messages: list[Message] | None,
    system: str | None,
    prompt: str | None,
    dedent: bool,
    max_retries: int,
    **kwargs,
) -> Any:
    collected_messages = _create_messages(
        messages=messages,
        system=system,
        prompt=prompt,
        dedent=dedent,
    )
    try:
        async for attempt in AsyncRetrying(
            stop=stop_after_attempt(max_retries),
            wait=wait_exponential(multiplier=1, min=1, max=10),
            reraise=True,
        ):
            with attempt:
                try:
                    async with semaphore:
                        await asyncio.sleep(random.uniform(0, 0.1))
                        return await generate(
                            model=model_name,
                            messages=collected_messages,
                            **kwargs,
                        )
                except Exception as e:
                    logger.debug(
                        f"Retry ({attempt.retry_state.attempt_number} of {max_retries}): {e}"
                    )
                    raise e
    except RetryError as e:
        logger.debug(f"Failed after {max_retries} attempts: {e}")
        raise e


def _create_messages(
    messages: list[Message] | None = None,
    system: str | None = None,
    prompt: str | None = None,
    dedent: bool = True,
) -> list[Message]:
    # check for invalid combinations
    if messages and (system or prompt):
        raise ValueError("Cannot specify both 'messages' and 'system'/'prompt'")

    # create messages if passed prompts
    if not messages:
        messages = []
        if system:
            messages.append(system_message(system))
        if prompt:
            messages.append(user_message(prompt))

    # apply dedent if needed
    return messages if not dedent else [m.dedent() for m in messages]



from typing import Any, Type, TypeVar

import anthropic
from pydantic import BaseModel

from agentlens import lens
from agentlens.provider import InferenceCost, Message, Provider

T = TypeVar("T", bound=BaseModel)


class Anthropic(Provider):
    # Cost per million tokens in USD
    MODEL_COSTS = {
        "claude-3-5-sonnet-20240620": (3.00, 15.00),
        "claude-3-opus-20240229": (15.00, 75.00),
        "claude-3-sonnet-20240229": (3.00, 15.00),
        "claude-3-haiku-20240307": (0.25, 1.25),
    }

    def __init__(
        self,
        api_key: str | None = None,
        max_connections: dict[str, int] | None = None,
        max_connections_default: int = 10,
    ):
        super().__init__(
            name="anthropic",
            max_connections=max_connections,
            max_connections_default=max_connections_default,
        )
        self.client = anthropic.AsyncAnthropic(api_key=api_key)

    def get_token_costs(self, model: str) -> tuple[float, float]:
        """Returns (input_cost_per_million, output_cost_per_million)"""
        return self.MODEL_COSTS.get(model, (0.0, 0.0))

    def _update_cost(
        self,
        model: str,
        response: Any,
        inference_cost: InferenceCost | None,
    ) -> None:
        cost = lens.run.cost
        usage = getattr(response, "usage", None)
        if usage:
            input_cost_per_token, output_cost_per_token = self.get_token_costs(model)
            cost.input += usage.input_tokens * input_cost_per_token / 1_000_000
            cost.output += usage.output_tokens * output_cost_per_token / 1_000_000

    async def generate_text(
        self,
        *,
        model: str,
        messages: list[Message],
        inference_cost: InferenceCost | None = None,
        **kwargs,
    ) -> str:
        completion = await self.client.messages.create(
            model=model,
            messages=[m.model_dump() for m in messages],  # type: ignore
            **kwargs,
        )

        self._update_cost(model, completion, inference_cost)
        assert completion.content is not None
        return completion.content[0].text

    async def generate_object(
        self,
        *,
        model: str,
        messages: list[Message],
        schema: Type[T],
        **kwargs,
    ) -> T:
        raise NotImplementedError("Anthropic does not support object generation")



import asyncio
import sys
import traceback
from importlib import import_module
from pathlib import Path

import typer
from typing_extensions import Annotated

app = typer.Typer()
run_app = typer.Typer()
app.add_typer(run_app, name="run")


@run_app.callback(invoke_without_command=True)
def run(
    file_path: Annotated[str, typer.Argument(help="Path to the Python file to run")],
    function_name: Annotated[str, typer.Argument(help="Name of the function to run")],
):
    """Run a Python function with AgentLens console visualization"""
    try:
        # Convert file path to module path
        path = Path(file_path)
        if not path.suffix == ".py":
            raise ValueError("File must be a Python file")

        # Convert path/to/file.py to path.to.file
        module_path = str(path.with_suffix("")).replace("/", ".").replace("\\", ".")

        # Import the module
        module = import_module(module_path)

        # Get the function
        func = module.__dict__.get(function_name)
        if func is None or not callable(func):
            available_functions = [name for name, item in module.__dict__.items() if callable(item)]
            typer.echo(
                f"\nFunction '{function_name}' not found or not callable. Available functions: {', '.join(available_functions)}",
                err=True,
            )
            raise typer.Exit(1)

        # Parse args into sys.argv for the function's CLI parser
        sys.argv = [file_path]

        if asyncio.iscoroutinefunction(func):
            asyncio.run(func())
        else:
            func()

    except ImportError:
        traceback.print_exc()
        raise typer.Exit(1)
    except Exception:
        traceback.print_exc()
        raise typer.Exit(1)


if __name__ == "__main__":
    app()
